#!/usr/bin/env python

import base64
import pickle
import signal
import subprocess
import sys
import threading
import time

# For reading the Data.FS in order to obtain affected object IDs from
# transaction IDs
import ZODB.FileStorage

from ..helpers import increment_txnid
from ..subcommand import SubCommand
from ..zodbsync import mod_read


class TreeOutdatedException(Exception):
    """Exception which is raised if the internal tree structure
    is not matching the actual Filesystem anymore"""

    pass


class Watch(SubCommand):
    """Periodically check for changes and record them"""

    # Connects to ZEO, builds a mirror of the tree structure of the objects,
    # periodically checks for new transactions, looks directly into the Data.FS
    # to get the object IDs affected by those transactions, and updates its
    # tree structure as well as the file system tree structure accordingly.

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--init",
            action="store_true",
            default=False,
            help="Internal mode for initialization subprocess",
        )

    def __init__(self, **kw):
        super(Watch, self).__init__(**kw)
        self.base_dir = self.sync.app_dir
        self.app = self.sync.app
        # an event that is fired if we are to be terminated
        self.exit = threading.Event()

        try:
            self.datafs_path = self.config["datafs_path"]
        except AttributeError:
            self.logger.exception("watch requires datafs_path in config")
            raise

        # mapping from object id to dict describing tree structure
        self.object_tree = {}

        # Mapping of additional object ids to OIDs of recorded objects. This is
        # currently only used for `User Folder`s which contain a
        # `PersistentMapping` and `User`s, which have their own OID and are not
        # children in the sense that they can be obtained using objectIds(). If
        # one of the additional OIDs is found to have been changed, the
        # original OID is assumed to have been changed instead.
        # If it turns out that it is possible to move a `User` from one `User
        # Folder` to another, they should instead be represented as separate
        # objects, but at least the management interface does not provide a
        # method for this.
        self.additional_oids = {}

    def _set_last_visible_txn(self):
        """Set self.last_visible_txn to a transaction ID such that every
        effect up to this ID is visible in the current transaction and every
        effect for transaction IDs above this are not yet visible.
        """
        self.last_visible_txn = self.app._p_jar._db.lastTransaction()

    def _store_last_visible_txn(self):
        """
        Store last visible transaction ID to disk if it changed.
        """
        if self.last_visible_txn != self.txnid_on_disk:
            self.txnid_on_disk = self.last_visible_txn
            self.sync.txn_write(base64.b64encode(self.last_visible_txn))

    def _init_tree(self, obj, parent_oid=None, path="/"):
        """Insert obj and everything below into self.object_tree."""
        if not hasattr(obj, "_p_oid"):
            # objects that have no oid are ignored
            return None
        # In some Python/Zope versions, _p_oid is a zodbpickle.binary, which is
        # not pickleable. We always convert it into bytes.
        oid = bytes(obj._p_oid)

        children = {}  # map oid -> id

        now = time.time()
        if self.last_report is None:
            self.last_report = now
        if now - self.last_report > 2:
            self.logger.info("Building tree: " + path)
            self.last_report = now

        self.object_tree[oid] = {
            "parent": parent_oid,
            "children": children,
            "path": path,
        }

        # If it turns out there are other objects needing such a hack, this
        # should probably be moved to object_types
        if obj.meta_type == "User Folder":
            self.additional_oids[bytes(obj.data._p_oid)] = oid
            for user in obj.getUsers():
                self.additional_oids[bytes(user._p_oid)] = oid

        for child_id, child_obj in sorted(obj.objectItems()):
            child_oid = self._init_tree(
                obj=child_obj, parent_oid=oid, path=path + child_id + "/"
            )
            if child_oid:
                children[child_oid] = child_id
        return oid

    def _read_changed_oids(self, txn_start, txn_stop):
        """
        Return a set of object IDS that are affected by the transactions with
        IDs between start and stop (incl.)
        """
        # FileIterator opens the Data.FS read-only and provides the following
        # fields and methods:
        # * _file: The file object for the opened Data.FS
        # * _read_data_header(pos): reads a data header at some position in the
        #   file
        # * _ltid: the transaction id that was last read.
        # It is also possible to iterate over FileIterator, which yields
        # transactions in the form of a TransactionRecord
        self.changed_oids = set()
        if txn_start > txn_stop:
            return
        storage = ZODB.FileStorage.FileIterator(
            self.datafs_path,
            start=txn_start,
            stop=txn_stop,
        )
        for txn in storage:
            # Each TransactionRecord has the following fields:
            # * _pos: Position of the first data header
            # * _tend: End of the last data block
            # * _file: A reference to the file
            pos = txn._pos
            while pos < txn._tend:
                dhead = storage._read_data_header(pos)
                # each data header has the fields and methods
                # * _recordlen(): gives the total length. Adding recordlen()
                #   advances to the next data header
                # * oid: the object ID
                # * plen: the size of the pickle data, which comes after the
                #   header
                dlen = dhead.recordlen()
                oid = self.additional_oids.get(bytes(dhead.oid), bytes(dhead.oid))
                self.changed_oids.add(oid)
                pos = pos + dlen

    def _remove_subtree(self, oid):
        """
        Remove a subtree from self.object_tree
        """
        todo = [oid]
        while todo:
            oid = todo.pop()
            todo.extend(self.object_tree[oid]["children"])
            del self.object_tree[oid]

    def _record_object(self, oid):
        """
        Store data of an object at the path stored in our object tree.
        """
        path = self.object_tree[oid]["path"]
        self.logger.info("Recording %s" % path)
        self.logger.debug("OID: " + repr(oid))

        obj = self.app._p_jar[oid]
        data = mod_read(obj=obj, default_owner=self.sync.default_owner)

        pathinfo = self.sync.fs_write(path=path, data=data)
        path_layer = pathinfo["layers"][pathinfo["layeridx"]]["ident"]
        current_layer = getattr(obj, "zodbsync_layer", None)
        if current_layer != path_layer:
            with self.sync.tm:
                obj.zodbsync_layer = path_layer

    def _update_objects(self):
        """
        Run through the changed oids and update the tree and the file system
        accordingly.
        """

        if not len(self.changed_oids):
            return
        self.logger.info("Found %s changed objects" % len(self.changed_oids))
        self.logger.debug("OIDs: " + str(sorted(self.changed_oids)))

        while len(self.changed_oids):
            # not all oids are part of our object tree yet, so we have to
            # iteratively update some at a time
            next_oids = self.changed_oids.intersection(self.object_tree.keys())
            if not len(next_oids):
                # The remaining oids are not reachable by any of the currently
                # existing nodes. This can happen during initialization, since
                # the tree structure is created as visible at the end of the
                # transaction chain and then affected objects are collected for
                # earlier transactions, but they might no longer exist
                break
            for oid in next_oids:
                if oid not in self.object_tree:
                    # Might have vanished from the tree in an earlier iteration
                    # because the parent was moved or renamed.
                    continue
                self._record_object(oid=oid)
                self._update_children(oid=oid)
                self.changed_oids.remove(oid)

        self.sync.fs_prune_empty_dirs()

    def _update_children(self, oid):
        """
        Check the current children of an object and compare with the stored
        children. Remove any superfluous children (oid not found or wrong id)
        and record any new children recursively.
        """
        obj = self.app._p_jar[oid]
        node = self.object_tree[oid]

        newchildren = {}
        for child_id, child_obj in obj.objectItems():
            if not hasattr(child_obj, "_p_oid"):
                continue
            newchildren[bytes(child_obj._p_oid)] = child_id

        # Go through old children and check if some are to be deleted
        for child_oid, child_id in list(node["children"].items()):
            if newchildren.get(child_oid) == child_id:
                continue
            self._remove_subtree(child_oid)
            del node["children"][child_oid]

        pathinfo = self.sync.fs_pathinfo(node["path"])
        self.sync.fs_prune(pathinfo, newchildren.values())

        # Add new children to changed_oids so they will also be recorded
        for child_oid, child_id in list(newchildren.items()):
            if child_oid in node["children"]:
                continue
            newpath = node["path"] + child_id + "/"

            if child_oid in self.object_tree:
                # The parent changed. Remove from there. Since the old parent
                # will also be changed, the call to fs_prune there will take
                # care of removing everything on the FS.
                old_parent = self.object_tree[child_oid]["parent"]
                del self.object_tree[old_parent]["children"][child_oid]
                self._remove_subtree(child_oid)

            self.changed_oids.add(child_oid)
            self.object_tree[child_oid] = {
                "parent": oid,
                "children": {},
                "path": newpath,
            }
            node["children"][child_oid] = child_id

    def quit(self, signo, _frame):
        """
        Signal handler
        """
        self.unregister_signals()
        self.exit.set()

    def register_signals(self):
        for sig in ("TERM", "HUP", "INT"):
            signal.signal(getattr(signal, "SIG" + sig), self.quit)

    def unregister_signals(self):
        for sig in ("TERM", "HUP", "INT"):
            signal.signal(getattr(signal, "SIG" + sig), signal.SIG_DFL)

    def setup(self):
        """
        Initially create tree and record anything that happened since the last
        running.
        """

        # During initialization, we report progress every 2 seconds.
        self.last_report = None

        # During normal operation, we always assume that the hard disk tree
        # structure is mirrored in object_tree, which mirrors the ZODB after
        # transaction A. When reading data in our Zope instance, we see the
        # ZODB after some later transaction B. We obtain the list of changed
        # object ids between A and B. Then we look up all objects that we know
        # of which were changed, record their meta data and move children
        # around, until our tree as well as the file system mirrors the state
        # at B (and our list of changed objects is empty).
        #
        # However, at startup the situation is different. Our object tree is
        # the same that we see through our Zope instance, which is the state
        # after B. The file system, on the other hand, mirrors the state after
        # A. Since we do not want to read the whole tree structure after A from
        # the file system (which would also require to store the OIDs), we do
        # not know which move operations would take us from A to B. Instead, we
        # collect a list of all changed paths and record them recursively.

        self.acquire_lock(timeout=300)
        self._set_last_visible_txn()
        self.sync.tm.begin()
        self._init_tree(self.app)

        # the transaction ID stored on disk is the last transaction whose
        # changes have already been recorded to disk. We increase it by one to
        # obtain all changes after that one
        self.txnid_on_disk = self.sync.txn_read()

        if self.txnid_on_disk is None:
            # no txnid found, record everything
            paths = ["/"]
        else:
            self.txnid_on_disk = base64.b64decode(self.txnid_on_disk)
            txn_start = increment_txnid(self.txnid_on_disk)

            # obtain all object ids affected by transactions between (the one
            # in last_txn + 1) and (the currently visible one) (incl.)
            self._read_changed_oids(txn_start=txn_start, txn_stop=self.last_visible_txn)
            paths = []
            while len(self.changed_oids):
                next_oids = self.changed_oids.intersection(self.object_tree.keys())
                if not len(next_oids):
                    # The remaining oids are not reachable by any of the
                    # currently existing nodes. This can happen during
                    # initialization since the tree structure is created as
                    # visible at the end of the transaction chain and then
                    # affected objects are collected for earlier transactions,
                    # but they might no longer exist
                    break
                paths.extend([self.object_tree[oid]["path"] for oid in next_oids])
                self.changed_oids.difference_update(next_oids)

        self.sync.record(paths)

        self.sync.tm.abort()

        # store an updated txnid on disk
        self._store_last_visible_txn()

        self.release_lock()

        self.logger.info("Setup complete")

    def spawned_setup(self):
        """
        Run setup in a separate process to reduce memory footprint of main
        process. During setup, the complete object tree is read into memory,
        which is not really released. Afterwards, the main process runs for a
        long time and uses much less memory. Therefore, we use a separate
        process.
        """
        cmd = [
            sys.executable,
            sys.argv[0],
            "--config",
            self.args.config,
            "watch",
            "--init",
        ]
        data = pickle.loads(subprocess.check_output(cmd))
        self.object_tree = data["tree"]
        self.additional_oids = data["add_oids"]
        self.last_visible_txn = self.txnid_on_disk = data["txn"]

    def dump_setup_data(self, stream=sys.stdout):
        """
        Print pickled setup data for usage in main process.
        """
        data = {
            "tree": self.object_tree,
            "add_oids": self.additional_oids,
            "txn": self.last_visible_txn,
        }
        # write binary to stdout - in Py3, this requires using
        # sys.stdout.buffer, in Py2 sys.stdout itself is used.
        pickle.dump(data, file=getattr(stream, "buffer", stream))

    def step(self):
        """Read new transactions, update the object tree and record all
        changes."""
        self.unregister_signals()
        self.acquire_lock(timeout=300)
        try:
            self.register_signals()

            start_txnid = increment_txnid(self.last_visible_txn)
            self._set_last_visible_txn()
            self._read_changed_oids(
                txn_start=start_txnid,
                txn_stop=self.last_visible_txn,
            )
            # make sure we see a consistent snapshot, even though we later
            # abort this transaction since we do not write anything
            self.sync.tm.begin()
            self._update_objects()
            self.sync.tm.abort()

            self._store_last_visible_txn()
        except TreeOutdatedException:
            self.logger.info("Exiting due to inconsistencies in filesystem")
            self.exit.set()
        finally:
            self.release_lock()

    def run(self, interval=10):
        """Setup and run in a loop."""
        if self.args.init:
            self.setup()
            self.dump_setup_data()
            return
        else:
            self.spawned_setup()
        while not self.exit.is_set():
            self.step()
            # a wait that is interrupted immediately if exit.set() is called
            self.exit.wait(interval)

        self.logger.info("Exited due to signal")



#!/usr/bin/env python

from ..subcommand import SubCommand


class Checkout(SubCommand):
    """Switch to another branch"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            default=False,
            help="Skip failed objects and continue",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Only check for conflicts and roll back at the end.",
        )
        parser.add_argument(
            "-b",
            action="store_true",
            default=False,
            help="Create branch.",
        )
        parser.add_argument(
            "-t", "--track", type=str, help="Set up upstream configuration."
        )
        parser.add_argument(
            "--reset",
            type=str,
            help="Reset branch onto given commit.",
        )
        parser.add_argument(
            "--rebase",
            type=str,
            help="Rebase branch onto given commit.",
        )
        parser.add_argument("branch", type=str, help="""Branch name""")

    @SubCommand.gitexec
    def run(self):
        self.logger.info("Checking out %s." % self.args.branch)
        cmd = ["checkout"]
        if self.args.b:
            cmd.append("-b")
        cmd.append(self.args.branch)
        if self.args.b and self.args.track:
            cmd.extend(["--track", self.args.track])
        self.gitcmd_run(*cmd)
        if self.args.reset:
            self.gitcmd_run("reset", "--hard", self.args.reset)
        if self.args.rebase:
            self.gitcmd_run("rebase", self.args.rebase)



#!/usr/bin/env python

import subprocess

from ..subcommand import SubCommand


class Exec(SubCommand):
    """Execute a command and play back any paths changed between old and new
    HEAD"""

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            default=False,
            help="Skip failed objects and continue",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Only check for conflicts and roll back at the end.",
        )
        parser.add_argument(
            "--nocd",
            action="store_true",
            default=False,
            help="Do not cd to git repo for command",
        )
        parser.add_argument("cmd", type=str, help="""command to be executed""")

    @SubCommand.gitexec
    def run(self):
        subprocess.check_call(
            self.args.cmd,
            cwd=None if self.args.nocd else self.sync.base_dir,
            shell=True,
        )



#!/usr/bin/env python

import json
from base64 import b64decode, b64encode

from .zodbsync import mod_read, mod_write


def launch(context, script, path, source=None, orig_source=None, encoding=None):
    """
    Launcher for external edit.

    If called without a source, it is used to create a control file that
    contains the authentication header found in the current request and the
    content of the file that can be found under path.

    If a source is provided, updates the object given by path, but only if its
    current source matches that given in orig_source.

    A wrapper script should be placed in the top-level of the ZODB that is only
    accessible for Manager and delegates to this.

    An encoding of None means the sources are Unicode. Other than that, only
    'base64' is supported, which means the sources are interpreted as base64
    encoded binary data.
    """

    resp = context.REQUEST.RESPONSE

    if source is None:
        content_type = "application/x-perfact-zopeedit"
        result = controlfile(
            context=context,
            path=path,
            url=script.absolute_url(),
        )
    else:
        content_type = "application/json"
        result = update(
            context=context,
            path=path,
            source=source,
            orig_source=orig_source,
            encoding=encoding,
        )
        result = json.dumps(result)

    resp.setHeader("Content-Type", content_type)
    return result


def read_obj(context, path, force_encoding=None):
    """
    Locate object at given path and return dictionary containing everything of
    interest.

    The 'source' field in the result is always text. If force_encoding is set
    to 'base64' or the source can not be interpreted as UTF-8, it is a base64
    representation of the actual source and the field 'encoding' is also set
    appropriately.
    """
    obj = context
    for part in path.split("/"):
        if not part:
            continue
        obj = getattr(obj, part)
    result = mod_read(obj)
    result["path"] = "/" + "/".join(obj.getPhysicalPath())
    result["parent"] = obj.aq_parent

    encoding = force_encoding
    if force_encoding and isinstance(result["source"], str):
        # We need bytes to encode with Base64
        result["source"] = result["source"].encode("utf-8")

    if not force_encoding and isinstance(result["source"], bytes):
        # Try to represent as UTF-8 for better readability.
        # If that does not work, switch to Base64.
        try:
            result["source"] = result["source"].decode("utf-8")
        except UnicodeDecodeError:
            encoding = "base64"

    if encoding == "base64":
        result["source"] = b64encode(result["source"]).decode("ascii")
        result["encoding"] = encoding

    return result


def controlfile(context, path, url):
    """
    Creates a control file that can be used by an external editor to update the
    contents of an object. The control file contains
    * the entrypoint url (which should be a script in Zope wrapping the
      launcher)
    * an authentication header
    * the path to the object in question
    * the meta_type of the object
    * the source of the object
    """

    data = read_obj(context, path)

    headers = [
        ("url", url),
        ("path", data["path"]),
        ("auth", context.REQUEST._auth),
        ("meta-type", data["type"]),
    ]
    encoding = data.get("encoding", None)
    if encoding:
        headers.append(("encoding", encoding))

    props = data.get("props", [])
    for prop in props:
        if ("id", "content_type") in prop:
            value = [pair for pair in prop if pair[0] == "value"]
            assert len(value), "Invalid property"
            headers.append(("content-type", value[0][1]))
            break

    result = (
        "".join(["{}: {}\n".format(*header) for header in headers])
        + "\n"
        + data["source"]
    )

    return result


def update(context, path, source, orig_source, encoding):
    """
    Update the object with the given source, but only if its current source
    matches the expected orig_source.

    If encoding is set to base64, the sources are considered to be base64
    encoded.
    """
    assert encoding in (None, "base64"), "Invalid encoding"

    try:
        data = read_obj(context, path, force_encoding=encoding)
    except AttributeError:
        return {"error": path + " not found"}

    if data["source"] != orig_source:
        return {"error": "Object was changed in the meantime. Please reload."}

    if encoding == "base64":
        data["source"] = b64decode(source)
    elif encoding is None:
        data["source"] = source
    obj_id = path.rstrip("/").rsplit("/", 1)[-1]
    mod_write(data, parent=data["parent"], obj_id=obj_id)

    return {"success": True}



#!/usr/bin/env python

import json
import os
import shutil
import subprocess
import sys

import filelock

from .helpers import Namespace


class SubCommand(Namespace):
    """
    Base class for different sub-commands to be used by zodbsync.
    """

    # The presence of one of these in the .git folder indicates that some
    # process was not finished correctly, which is used to trigger a rollback
    # in some operations. Are these all?
    git_state_indicators = [
        "rebase-merge",
        "rebase-apply",
        "CHERRY_PICK_HEAD",
        "MERGE_HEAD",
        "REVERT_HEAD",
    ]

    @staticmethod
    def add_args(parser):
        """Overwrite to add arguments specific to sub-command."""
        pass

    def acquire_lock(self, timeout=10):
        if self.args.no_lock:
            return
        try:
            self.lock.acquire(timeout=1)
        except filelock.Timeout:
            self.logger.debug("Acquiring exclusive lock...")
            try:
                self.lock.acquire(timeout=timeout - 1)
            except filelock.Timeout:
                self.logger.error("Unable to acquire lock.")
                sys.exit(1)

    def release_lock(self):
        if not self.args.no_lock:
            self.lock.release()

    @staticmethod
    def with_lock(func):
        """
        Decorator for instance methods that are enveloped by a lock
        """

        def wrapper(self, *args, **kwargs):
            self.acquire_lock()
            try:
                result = func(self, *args, **kwargs)
            finally:
                self.release_lock()
            return result

        return wrapper

    def gitcmd(self, *args):
        # use "--no-pager" instead of "-P" for compatibility / readability
        return ["git", "--no-pager", "-C", self.config["base_dir"]] + list(args)

    def gitcmd_run(self, *args):
        """Wrapper to run a git command."""
        subprocess.check_call(self.gitcmd(*args))

    def gitcmd_try(self, *args):
        """Wrapper to run a git command, returning return code."""
        return subprocess.call(self.gitcmd(*args))

    def gitcmd_output(self, *args):
        """Wrapper to run a git command and return the output."""
        return subprocess.check_output(self.gitcmd(*args), universal_newlines=True)

    def datafs_filesystem_path(self, path):
        """Create absolute filesystem path from Data.fs path"""

        if path.startswith("./"):
            path = path[2:]

        if path.startswith("/"):
            path = path[1:]

        data_fs_path = path
        if path.startswith(self.sync.site):
            filesystem_path = os.path.join(self.config["base_dir"], path)
        else:
            filesystem_path = os.path.join(
                self.config["base_dir"], self.sync.site, path
            )

        return data_fs_path, filesystem_path

    @staticmethod
    def unpack_source(src, tgt):
        """
        Unpack files from source to target.
        Each non-hidden folder in src is copied over, each non-hidden file is
        unpacked, both removing superfluous files in the target.
        """
        targetitems = []
        srcitems = os.listdir(src)
        for entry in srcitems:
            if entry.startswith("."):
                continue
            path = f"{src}/{entry}"
            if os.path.isdir(path):
                # p.e. __root__ or __schema__ as folders
                # Sometimes, there might be some residual folder with .dpkg-new
                # files or similar, even though this is now supplied as file.
                other = [
                    other
                    for other in srcitems
                    if other.startswith(entry) and other != entry
                ]
                if other:
                    continue
                targetitems.append(entry)
                cmd = ["rsync", "-a", "--delete-during", f"{path}/", f"{tgt}/{entry}/"]
            else:
                # p.e. __root__.tar.gz -> Unpack to __root__/
                basename = entry.split(".")[0]
                targetitems.append(basename)
                os.makedirs(f"{tgt}/{basename}", exist_ok=True)
                cmd = [
                    "tar",
                    "xf",
                    path,
                    "-C",
                    f"{tgt}/{basename}/",
                    "--recursive-unlink",
                    "-m",
                ]
            subprocess.run(cmd, check=True)
        for entry in os.listdir(tgt):
            if entry.startswith(".") or entry in targetitems:
                continue
            shutil.rmtree(f"{tgt}/{entry}")

    def _branch_info(self):
        """Returns currently checked out branch as well as where each branch
        points."""
        current = self.gitcmd_output("rev-parse", "--abbrev-ref", "HEAD").strip()

        branches = {}  # branchname -> commitid
        output = self.gitcmd_output("show-ref", "--heads")
        for line in output.strip().split("\n"):
            commit, refname = line.split()
            refname = refname[len("refs/heads/") :]
            branches[refname] = commit

        return (current, branches)

    def check_repo(self):
        """Check for unstaged changes and memorize current commit. Move
        unstaged changes away via git stash"""
        self.unstaged_changes = [
            line[3:]
            for line in self.gitcmd_output("status", "--untracked-files", "-z").split(
                "\0"
            )
            if line
        ]

        self.orig_branch, self.branches = self._branch_info()

        if self.unstaged_changes:
            self.logger.warning("Unstaged changes found. Moving them out of the way.")
            self.gitcmd_run("stash", "push", "--include-untracked")

        # The commit to compare to with regards to changed files
        self.orig_commit = self.branches[self.orig_branch]

    def _playback_paths(self, paths):
        paths = self.sync.prepare_paths(paths)
        dryrun = self.args.dry_run

        playback_hook = self.config.get("playback_hook", None)
        if playback_hook and os.path.isfile(playback_hook):
            proc = subprocess.Popen(
                playback_hook,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
            out, _ = proc.communicate(json.dumps({"paths": paths}))
            returncode = proc.returncode
            if returncode:
                raise AssertionError(
                    "Error calling playback hook, returncode {}, [[{}]] on {}".format(
                        returncode, playback_hook, out
                    )
                )
            phases = json.loads(out)
        else:
            phases = [{"name": "playback", "paths": paths}]
            if self.config.get("run_after_playback", None):
                phases[-1]["cmd"] = self.config["run_after_playback"]

        for ix, phase in enumerate(phases):
            phase_name = phase.get("name") or str(ix)
            phase_cmd = phase.get("cmd")

            self.sync.playback_paths(
                paths=phase["paths"],
                recurse=False,
                override=True,
                skip_errors=self.args.skip_errors,
                dryrun=dryrun,
            )

            if dryrun or not (phase_cmd and os.path.isfile(phase_cmd)):
                continue

            self.logger.info("Calling phase %s, command: %s", phase_name, phase_cmd)
            proc = subprocess.Popen(
                phase_cmd,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
            out, _ = proc.communicate(json.dumps({"paths": phase["paths"]}))
            returncode = proc.returncode

            if returncode:
                self.logger.error("Error during phase command %s, %s", returncode, out)
                if sys.stdin.isatty():
                    print("Enter 'y' to continue, other to rollback")
                    res = input()
                    if res == "y":
                        continue

                raise AssertionError("Unrecoverable error in phase command")
            else:
                self.logger.info(out)

    @staticmethod
    def gitexec(func):
        """
        Decorator for wrapping an operation with playback:
        - Stash unstaged changes away
        - memorize the current commit
        - do something
        - check for conflicts
        - play back changed objects (diff between old and new HEAD)
        - unstash
        """

        @SubCommand.with_lock
        def wrapper(self, *args, **kwargs):
            # Check for unstaged changes
            self.check_repo()

            try:
                self.paths = []
                func(self, *args, **kwargs)

                # Fail and roll back for any of the markers of an interrupted
                # git process (merge/rebase/cherry-pick/etc.)
                for fname in self.git_state_indicators:
                    path = os.path.join(self.sync.base_dir, ".git", fname)
                    assert not os.path.exists(path), "Git state not clean"

                files = {
                    line
                    for line in self.gitcmd_output(
                        "diff", self.orig_commit, "--name-only", "--no-renames"
                    )
                    .strip()
                    .split("\n")
                    if line
                }
                conflicts = files & set(self.unstaged_changes)
                assert not conflicts, "Change in unstaged files, aborting"

                # Make unique and sort
                self.paths = sorted(
                    {file for file in files if file.startswith(self.sync.site)}
                )

                self._playback_paths(self.paths)

                if self.args.dry_run:
                    self.abort()
                elif self.unstaged_changes:
                    self.gitcmd_run("stash", "pop")

            except Exception:
                self.logger.error("Error during operation. Resetting.")

                # Special handling in case of interrupted cherry-pick: show
                # differences in affected files
                cpfname = os.path.join(self.sync.base_dir, ".git/CHERRY_PICK_HEAD")
                if os.path.exists(cpfname):
                    with open(cpfname) as f:
                        failed_commit = f.read().strip()
                    output = self.gitcmd_output(
                        "diff-tree",
                        "--no-commit-id",
                        "--name-only",
                        "-r",
                        failed_commit,
                    )
                    affected_files = [
                        line for line in output.strip().split("\n") if line
                    ]
                    self.logger.error(
                        "The cherry-pick failed due to the following difference:"
                    )
                    try:
                        self.gitcmd_run(
                            "diff", failed_commit + "~", "HEAD", "--", *affected_files
                        )
                    except subprocess.CalledProcessError:
                        # Make sure the call to abort is still done, even if
                        # for example the list of affected_files is too long
                        self.logger.exception("Unable to show diff")

                self.abort()
                # if we are not in dryrun we can't be sure we havent already
                # committed some stuff to the data-fs so playback all paths
                # abort
                if not self.args.dry_run and self.paths:
                    self.sync.playback_paths(
                        paths=self.paths,
                        recurse=False,
                        override=True,
                        skip_errors=True,
                        dryrun=False,
                    )
                raise

            if not self.args.dry_run:
                is_ancestor = (
                    self.gitcmd_try(
                        "merge-base", "--is-ancestor", self.orig_commit, "HEAD"
                    )
                    == 0
                )
                if is_ancestor:
                    merge_commits = self.gitcmd_output(
                        "log",
                        "--oneline",
                        "--min-parents=2",
                        f"{self.orig_commit}..HEAD",
                    ).strip()
                    if not merge_commits:
                        head_commit = self.gitcmd_output("rev-parse", "HEAD").strip()
                        cmd = (
                            f'zodbsync exec "git revert '
                            f'{self.orig_commit}..{head_commit}"'
                        )
                        self.logger.info(
                            "Prepared Command for Rollback:\n%s",
                            cmd,
                        )

        return wrapper

    def create_file(self, file_path, content, binary=False):
        flags = "wb" if binary else "w"
        with open(file_path, flags) as create_file:
            create_file.write(content)

    def abort(self):
        """Abort actions on repo and revert stash. check_repo must be
        called before this can be used"""
        current, branches = self._branch_info()
        # reset currently checked out branch
        target = self.branches.get(current)
        if target is None:
            # The branch was not originally present - we still need to reset it
            # to abort any operation
            target = branches[current]
        self.gitcmd_run("reset", "--hard", target)

        # reset all other branches
        for branch in self.branches:
            if branch == current:
                continue
            if branches[branch] == self.branches[branch]:
                continue
            self.gitcmd_run("branch", "-f", branch, self.branches[branch])

        # check out original branch
        if current != self.orig_branch:
            self.gitcmd_run("checkout", self.orig_branch)

        if self.unstaged_changes:
            self.gitcmd_run("stash", "pop")

    def run(self):
        """
        Overwrite for the action that is to be performed if this subcommand is
        chosen.
        """
        print(self.args)



#!/usr/bin/env python

from ..subcommand import SubCommand


class FF(SubCommand):
    """
    Perform a fast-forward merge to the target commit and apply changed paths
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            default=False,
            help="Skip failed objects and continue",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Only check for conflicts and roll back at the end.",
        )
        parser.add_argument("commit", type=str, help="""Target commit""")

    @SubCommand.gitexec
    def run(self):
        target = self.args.commit
        self.logger.info("Attempting fast-forward merge to %s." % target)
        self.gitcmd_run("merge", "--ff-only", target)



#!/usr/bin/env python

import subprocess

from ..subcommand import SubCommand


class WithLock(SubCommand):
    """Execute a shell command by first grabbing the lock"""

    subcommand = "with-lock"
    connect = False

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "cmd",
            type=str,
            help="Shell-command to be executed",
        )

    @SubCommand.with_lock
    def run(self):
        subprocess.check_call(self.args.cmd, shell=True)



from .extedit import launch as extedit_launch
from .helpers import db_modtime, obj_modtime
from .zodbsync import mod_read, mod_write

__all__ = [
    "mod_read",
    "mod_write",
    "obj_modtime",
    "db_modtime",
    "extedit_launch",
]



# -*- coding: utf-8 -*-
import pytest

from .. import helpers


def test_remove_redundant_paths():
    """
    Check that redundant subpaths are actually removed
    """
    paths = [
        "/test",
        "/test/sub",
        "/another",
    ]
    target = [
        "/another",
        "/test",
    ]
    helpers.remove_redundant_paths(paths)
    assert paths == target


def test_remove_redundant_paths_only_real_subpaths():
    """
    Check that paths are only recognized as redundant if they are actually
    subpaths, not if the last path component starts with the other.
    """
    paths = ["/test", "/test2"]
    new_paths = paths[:]
    helpers.remove_redundant_paths(new_paths)
    assert paths == new_paths


def test_converters():
    """
    Several tests for to_* methods
    """
    for value in ["test", b"test"]:
        assert helpers.to_bytes(value) == b"test"
        assert helpers.to_string(value) == "test"
    assert helpers.to_string([1]) == "[1]"
    assert helpers.to_bytes([1]) == b"[1]"
    assert helpers.to_bytes(memoryview(b"test")) == b"test"


def test_StrRepr():
    """
    Check recursive version of str_repr with a typical configuration for what
    is split to occupy one line for each element, reproducing the shown
    formatting.
    """
    fmt = (
        """
[
    ('content', [
        'a',
        'b',
    ]),
    ('owner', (['acl_users'], 'admin')),
    ('perms', [
        ('View', False, [
            'Role_1',
            'Role_2',
        ]),
    ]),
    ('props', [
        [('id', 'columns'), ('type', 'tokens'), ('value', (
            'a',
            'b',
            'c',
        ))],
        [('id', 'other'), ('type', 'lines'), ('value', (
            'x',
            'y',
            'z',
        ))],
        [('id', 'scalar'), ('type', 'string'), ('value', 'test')],
    ]),
]
    """.strip()
        + "\n"
    )

    data = dict(helpers.literal_eval(fmt))
    rules = {
        "perms": [4],
        "props": [5],
    }
    assert fmt == helpers.StrRepr()(data, rules)


def test_StrReprLegacy():
    """
    Reproduce the shown formatting of StrRepr when using legacy mode
    """
    fmt = (
        """
[
    ('content', [
        'a',
        'b',
        ]),
    ('owner', (['acl_users'], 'admin')),
    ('perms', [('View', False, ['Role_1', 'Role_2'])]),
    ('props', [
        [('id', 'columns'), ('type', 'tokens'), ('value', ('a', 'b', 'c'))],
        [('id', 'other'), ('type', 'lines'), ('value', ('x', 'y', 'z'))],
        [('id', 'scalar'), ('type', 'string'), ('value', 'test')],
        ]),
]
    """.strip()
        + "\n"
    )
    data = dict(helpers.literal_eval(fmt))
    assert fmt == helpers.StrRepr()(data, legacy=True)


def test_literal_eval():
    tests = [
        ["b'test'", b"test"],
        ["{1: 2}", {1: 2}],
        ["[1, 2, 3]", [1, 2, 3]],
        ["None", None],
    ]
    for orig, compare in tests:
        assert helpers.literal_eval(orig) == compare
    assert helpers.literal_eval("1 + 2") == 3
    assert helpers.literal_eval("-True") == -1
    with pytest.raises(Exception):
        helpers.literal_eval("f(1)")


def test_path_diff():
    """Check that path_diff also handles cases where the last element is not
    the same in both lists."""
    old = [
        ("Abc", "1234"),
        ("Def", "afaf"),
        ("Xyz", "yzyz"),
    ]
    new = [
        ("Abc", "1234"),
        ("Def", "axax"),
        ("Yyy", "yzyz"),
    ]
    result = helpers.path_diff(old, new)
    assert result == {"Def", "Xyz", "Yyy"}



#!/usr/bin/env python
from . import helpers
from .object_mixins import MixinModObj


class ModObj(object):
    meta_types = []

    @staticmethod
    def create(obj, data, obj_id):
        return

    @staticmethod
    def read(obj):
        return ()

    @staticmethod
    def write(obj, data):
        return

    @classmethod
    def collect_handlers(cls):
        """
        Create a dictionary mapping the supported meta_types of each class to
        the handler class by recursing into subclasses.
        """
        result = {meta_type: cls for meta_type in cls.meta_types}
        for sub in cls.__subclasses__():
            result.update(sub.collect_handlers())
        return result


class UserFolderObj(ModObj):
    meta_types = [
        "User Folder",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addUserFolder()

    @staticmethod
    def read(obj):
        users = []
        for user in obj.getUsers():
            users.append(
                (
                    user.getUserName(),
                    user._getPassword(),
                    user.roles,
                    user.getDomains(),
                )
            )
        return {"users": users}

    @staticmethod
    def write(obj, data):
        users = obj.getUsers()
        current_users = [user.getUserName() for user in users]
        target_users = [user[0] for user in data["users"]]
        obj._doDelUsers([u for u in current_users if u not in target_users])
        for user in data["users"]:
            # according to AccessControl/userfolder.py, an existing user of the
            # same name is simply overwritten by _doAddUser
            obj._doAddUser(
                user[0],  # username
                "",  # password is set separately
                user[2],  # roles
                user[3],  # domains
            )
            # _doAddUser encrypts the given password, but the password in the
            # dump is already encrypted, so we have to set it manually
            obj.getUserById(user[0]).__ = user[1]


class DTMLDocumentObj(ModObj):
    meta_types = [
        "DTML Document",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addDTMLDocument(id=obj_id)

    @staticmethod
    def read(obj):
        return {"source": helpers.to_bytes(obj.raw)}

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            data=helpers.to_string(data["source"]),
            title=data["title"],
        )


class DTMLMethodObj(DTMLDocumentObj):
    meta_types = [
        "DTML Method",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addDTMLMethod(id=obj_id)


class DTMLTeXObj(DTMLDocumentObj):
    meta_types = [
        "DTML TeX",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["DTMLTeX"].manage_addDTMLTeX(id=obj_id)


class ZForceObj(ModObj):
    meta_types = [
        "ZForce",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["ZForce"].manage_addZForce(
            id=obj_id,
            title="",
            query_id="",
            fields_id="",
        )

    @staticmethod
    def read(obj):
        return {}

    @staticmethod
    def write(obj, data):
        obj.manage_changeProperties(title=data["title"])


class ZSQLMethodObj(ModObj):
    meta_types = [
        "Z SQL Method",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["ZSQLMethods"].manage_addZSQLMethod(
            id=obj_id,
            title=data["title"],
            connection_id=data["connection_id"],
            arguments=data["args"],
            template=helpers.to_string(data["source"]),
        )

    @staticmethod
    def read(obj):
        return {
            "args": obj.arguments_src,
            "connection_id": obj.connection_id,
            "source": helpers.to_bytes(obj.src),
            "advanced": sorted(
                [
                    ("connection_hook", obj.connection_hook),
                    ("max_rows", obj.max_rows_),
                    ("max_cache", obj.max_cache_),
                    ("cache_time", obj.cache_time_),
                    ("class_name", obj.class_name_),
                    ("class_file", obj.class_file_),
                ]
            ),
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data["title"],
            connection_id=data["connection_id"],
            arguments=data["args"],
            template=helpers.to_string(data["source"]),
        )

        # Advanced settings
        adv = dict(data["advanced"])
        obj.manage_advanced(**adv)


class ExternalMethodObj(ModObj):
    meta_types = [
        "External Method",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["ExternalMethod"].manage_addExternalMethod(
            id=obj_id,
            title=data["title"],
            module=data["module"],
            function=data["function"],
        )

    @staticmethod
    def read(obj):
        return {
            "function": obj.function(),
            "module": obj.module(),
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data["title"],
            module=data["module"],
            function=data["function"],
        )


class FileObj(ModObj):
    meta_types = [
        "File",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addFile(id=obj_id)

    @staticmethod
    def read(obj):
        # XXX Precondition
        # Read chunked source from File/Image objects.
        return {"source": helpers.read_pdata(obj)}

    @staticmethod
    def write(obj, data):
        pd = helpers.prop_dict(data)

        # XXX Precondition?
        obj.manage_edit(
            filedata=data["source"],
            content_type=pd["content_type"],
            title=data["title"],
        )


class ImageObj(FileObj):
    meta_types = [
        "Image",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addImage(id=obj_id, file="")


class FolderObj(ModObj):
    meta_types = [
        "Folder",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addFolder(id=obj_id)

    @staticmethod
    def read(obj):
        # Site Access
        try:
            get_ar = obj.manage_addProduct["SiteAccess"].manage_getAccessRule
        except (AttributeError, KeyError):
            get_ar = None
        if get_ar:
            accessrule = get_ar and get_ar()
            if accessrule:
                return {"accessrule": accessrule}

        return {}

    @staticmethod
    def write(obj, data):
        obj.manage_changeProperties(title=data["title"])

        # Access Rule
        accessrule = data.get("accessrule", None)
        if accessrule:
            obj.manage_addProduct["SiteAccess"].manage_addAccessRule(
                method_id=accessrule
            )


class FolderOrderedObj(FolderObj):
    meta_types = [
        "Folder (Ordered)",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["OFSP"].manage_addOrderedFolder(id=obj_id)

    @staticmethod
    def read(obj):
        result = FolderObj.read(obj)
        # ordered folders store their contents to represent the ordering
        result["contents"] = [a[0] for a in obj.objectItems()]
        return result

    @staticmethod
    def write(obj, data):
        obj.manage_changeProperties(title=data["title"])

        # Access Rule
        accessrule = data.get("accessrule", None)
        if accessrule:
            obj.manage_addProduct["SiteAccess"].manage_addAccessRule(
                method_id=accessrule
            )

    @staticmethod
    def fix_order(obj, data):
        # sort children for ordered folders
        contents = data.get("contents", [])
        srv_contents = [a[0] for a in obj.objectItems()]

        # only use contents that are present in the object
        contents = [a for a in contents if a in srv_contents]
        obj.moveObjectsByDelta(contents, -len(contents))


class PageTemplateObj(ModObj):
    meta_types = [
        "Page Template",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["PageTemplates"].manage_addPageTemplate(
            id=obj_id, text=""
        )

    @staticmethod
    def read(obj):
        return {"source": obj._text}

    @staticmethod
    def write(obj, data):
        obj.pt_setTitle(data["title"], "utf-8")
        obj.write(data["source"])


class RAMCacheManagerObj(ModObj):
    meta_types = [
        "RAM Cache Manager",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["StandardCacheManagers"].manage_addRAMCacheManager(
            id=obj_id
        )

    @staticmethod
    def read(obj):
        return {"settings": sorted(obj.getSettings().items())}

    @staticmethod
    def write(obj, data):
        obj.manage_editProps(
            title=data["title"],
            settings=dict(data["settings"]),
        )


class AcceleratedHTTPCacheManagerObj(RAMCacheManagerObj):
    meta_types = [
        "Accelerated HTTP Cache Manager",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct[
            "StandardCacheManagers"
        ].manage_addAcceleratedHTTPCacheManager(id=obj_id)


class ScriptPythonObj(ModObj):
    meta_types = [
        "Script (Python)",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["PythonScripts"].manage_addPythonScript(id=obj_id)

    @staticmethod
    def read(obj):
        return {
            "bindings": sorted(obj.getBindingAssignments().getAssignedNames().items()),
            "args": obj.params(),
            "source": helpers.to_bytes(obj.body()),
            "proxy_roles": sorted(list(obj._proxy_roles)),
        }

    @staticmethod
    def write(obj, data):
        obj.ZPythonScript_setTitle(title=data["title"])
        obj.ZPythonScript_edit(
            params=data["args"], body=helpers.to_string(data["source"])
        )
        obj.ZBindings_edit(mapping=dict(data["bindings"]))
        obj.manage_proxy(roles=data["proxy_roles"])


class ZPsycopgDAObj(ModObj):
    meta_types = [
        "Z Psycopg 2 Database Connection",
        "Z Psycopg Database Connection",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct["ZPsycopgDA"].manage_addZPsycopgConnection(
            id=obj_id,
            title=data["title"],
            connection_string=data["connection_string"],
        )

    @staticmethod
    def read(obj):
        # late additions may not yet be everywhere in the Data.fs
        obj_dict = {
            "autocommit": getattr(obj, "autocommit", False),
            "readonlymode": getattr(obj, "readonlymode", False),
            "connection_string": obj.connection_string,
            "encoding": obj.encoding,
            "tilevel": obj.tilevel,
            "zdatetime": obj.zdatetime,
        }
        # Place additional parameters into the object dict only if
        # they're set to non-default values
        if hasattr(obj, "use_tpc") and obj.use_tpc:
            obj_dict["use_tpc"] = obj.use_tpc
        if hasattr(obj, "datetime_str") and obj.datetime_str:
            obj_dict["datetime_str"] = obj.datetime_str
        return obj_dict

    @staticmethod
    def write(obj, data):
        parameters = {
            "title": data["title"],
            "connection_string": data["connection_string"],
            "zdatetime": data["zdatetime"],
            "tilevel": data["tilevel"],
            "autocommit": data.get("autocommit", False),
            "readonlymode": data.get("readonlymode", False),
            "encoding": data["encoding"],
        }
        if hasattr(obj, "use_tpc"):
            parameters["use_tpc"] = data.get("use_tpc", False)
        if hasattr(obj, "datetime_str"):
            parameters["datetime_str"] = data.get("datetime_str", False)
        obj.manage_edit(**parameters)


class ZPyODBCDAObj(ModObj):
    meta_types = [
        "Z PyODBC Database Connection",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct["ZPyODBCDA"].addpyodbcConnectionBrowser(
            id=obj_id,
            title=data["title"],
            connection_string=data["connection_string"],
            auto_commit=data["autocommit"],
            MaxRows=data["maxrows"],
        )

    @staticmethod
    def read(obj):
        return {
            "autocommit": obj.auto_commit,
            "connection_string": obj.connx_string,
            "maxrows": obj.MaxRows,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data["title"],
            connection_string=data["connection_string"],
            auto_commit=data["autocommit"],
            MaxRows=data["maxrows"],
        )


class ZcxOracleDAObj(ModObj):
    meta_types = [
        "Z cxOracle Database Connection",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct["ZcxOracleDA"].manage_addZcxOracleConnection(
            id=obj_id,
            title=data["title"],
            connection_string=data["connection_string"],
        )

    @staticmethod
    def read(obj):
        return {
            "connection_string": obj.connection_string,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data["title"],
            connection_string=data["connection_string"],
        )


class ZsapdbDAObj(ModObj):
    meta_types = [
        "Z sap Database Connection",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct["ZsapdbDA"].manage_addZsapdbConnection(
            id=obj_id,
            title=data["title"],
            connection_string=data["connection_string"],
        )

    @staticmethod
    def read(obj):
        return {
            "connection_string": obj.connection_string,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data["title"],
            connection_string=data["connection_string"],
        )


class SimpleUserFolderObj(FolderObj):
    meta_types = [
        "Simple User Folder",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct["SimpleUserFolder"].addSimpleUserFolder()


class MailHostObj(ModObj):
    meta_types = [
        "Mail Host",
    ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct["MailHost"].manage_addMailHost(
            id=obj_id,
            title=data["title"],
            smtp_host=data["smtp_host"],
            smtp_port=data["smtp_port"],
        )

    @staticmethod
    def read(obj):
        return {
            "smtp_host": obj.smtp_host,
            "smtp_port": obj.smtp_port,
            "smtp_uid": obj.smtp_uid,
            "smtp_pwd": obj.smtp_pwd,
            "force_tls": obj.force_tls,
            "smtp_queue": obj.smtp_queue,
            "smtp_queue_directory": obj.smtp_queue_directory,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_makeChanges(
            title=data["title"],
            smtp_host=data["smtp_host"],
            smtp_port=data["smtp_port"],
            smtp_uid=data.get("smtp_uid", ""),
            smtp_pwd=data.get("smtp_pwd", ""),
            force_tls=data.get("force_tls", False),
            smtp_queue=data.get("smtp_queue", ""),
            smtp_queue_directory=data.get("smtp_queue_directory", ""),
        )


object_handlers = ModObj.collect_handlers()


def mod_implemented_handlers(obj, meta_type):
    """
    Return the handlers of interfaces that the object implements. This always
    includes the one defined from the meta_type and may also include some of
    the mixins.
    """
    return [object_handlers[meta_type]] + [
        cls for cls in MixinModObj.__subclasses__() if cls.implements(obj)
    ]



import argparse

from perfact.zodbsync.main import Runner

try:
    # psql dump for backwards compatibility
    import perfact.dbbackup
except ImportError:
    pass


def zodbsync():
    Runner().run()


def zoperecord():
    parser = argparse.ArgumentParser(
        description="Record the Data.fs",
        epilog="""This script is deprecated in favor of zodbsync. Only bare
        functionality is provided for backwards compatibility with existing
        cron entries.
        """,
    )
    parser.add_argument(
        "--lasttxn",
        action="store_true",
        default=False,
        help="Record only transactions since the last used.",
    )

    args = parser.parse_args()
    if args.lasttxn:
        cmd = "record --lasttxn"
    else:
        cmd = "record --commit /"

    runner = Runner().parse(*cmd.split())

    # dump tables and schemas if run without --lasttxn and if the corresponding
    # variables are found in the config - this is only for backwards
    # compatibility, this should be done by perfact-dbrecord instead.
    config = runner.config
    databases = getattr(config, "databases", None)
    if not args.lasttxn and databases is not None:
        runner.logger.warn(
            "Deprecation warning: dumping PostgreSQL schema and tables, which"
            " should be done by perfact-dbrecord instead."
        )
        msgbak = config.commit_message
        config.commit_message += " (Database)"
        perfact.dbbackup.git_snapshot(config)
        config.commit_message = msgbak

    runner.run()


def zopeplayback():
    assert False, (
        "perfact-zopeplayback is no longer supported as executable, please use"
        " zodbsync instead."
    )



# Path of the Zope instance configuration to use to instantiate the application
# object
conf_path = "/var/lib/zope4/ema/etc/zope.conf"

# Path to Data.fs which is needed for lookup of object IDs from transaction IDs
# with zodbsync watch
datafs_path = "/var/lib/zope4/zeo/var/Data.fs"

# user that is used to create commits
manager_user = "perfact"

# create the manager user with a default password if not present
create_manager_user = True

# sets the default owner for objects that have no owner in the file system
# representation
default_owner = "perfact"

# use default owner even if we're told otherwise by meta file
force_default_owner = False

# Base directory of the repository
base_dir = "/opt/perfact/dbutils-zoperepo"

# default settings for git repos
commit_name = "Zope Developer"
commit_email = "zope-devel@example.de"
commit_message = "Generic commit message."

# email address to send commit summaries of default commits to
# codechange_mail = "zope-devel@example.de"
# codechange_sender = "no-reply-zodbsync-changes@example.de"

# Path to script which is called to define the phases of playback to be
# executed.
# playback_hook = '/usr/share/perfact/zope4-tools/zodbsync-playback-hook'

# Path to script that is called for postprocessing after a playback if it exists
# run_after_playback = '/usr/share/perfact/zope4-tools/zodbsync-postproc'



#!/usr/bin/env python

import subprocess as sp

from ..subcommand import SubCommand


class LayerInit(SubCommand):
    """Register layers from source to work_dir, assuming objects are already in
    the Data.FS, but are now to be provided by a new layer."""

    subcommand = "layer-init"

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "ident",
            type=str,
            nargs="*",
            help="Layer identifier(s). May be * for all",
        )

    @SubCommand.with_lock
    def run(self):
        layers = {layer["ident"]: layer for layer in self.sync.layers if layer["ident"]}
        idents = self.args.ident
        if idents == ["*"]:
            idents = layers.keys()
        for ident in idents:
            assert ident in layers, "Invalid ident"
        for ident in idents:
            layer = layers[ident]
            source = layer["source"]
            target = layer["workdir"]
            self.unpack_source(source, target)
            sp.run(["git", "add", "."], cwd=target)
            sp.run(["git", "commit", "-m", "zodbsync layer-init"], cwd=target)



#!/usr/bin/env python

import os
import subprocess as sp

from ..subcommand import SubCommand


class LayerUpdate(SubCommand):
    """Update layers."""

    subcommand = "layer-update"

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Only check for conflicts and roll back at the end.",
        )
        parser.add_argument(
            "--skip-errors",
            action="store_true",
            help="Skip failed objects and continue",
            default=False,
        )
        parser.add_argument(
            "--message",
            "-m",
            type=str,
            default="zodbsync layer-update",
            help="Commit message base",
        )
        parser.add_argument(
            "ident",
            type=str,
            nargs="*",
            help="Layer identifier(s). May be * for all",
        )

    def commit_all(self, target, msg):
        """Commit all unstaged changes in target, returning the commit ID or
        None if there is no change."""
        sp.run(["git", "add", "."], cwd=target)
        if sp.run(["git", "commit", "-m", msg], cwd=target).returncode == 0:
            return sp.check_output(
                ["git", "rev-parse", "HEAD"], cwd=target, text=True
            ).strip()

    def run_layer(self, layer):
        """
        For given layer, commit any unstaged changes, update work_dir from
        source, commit that and play back any changes.
        """
        source = layer["source"]
        target = layer["workdir"]
        msg = self.args.message
        precommit = self.commit_all(target, f"{msg} (pre)")
        self.unpack_source(source, target)
        changes = [
            line[3:]
            for line in sp.check_output(
                ["git", "status", "--porcelain", "-u", "--no-renames"],
                cwd=target,
                text=True,
            ).split("\n")
            if line
        ]
        commit = None
        if changes:
            commit = self.commit_all(target, msg)
        self.restore[layer["ident"]] = (precommit, commit)
        return {
            os.path.dirname(line[len("__root__") :])
            for line in changes
            if line.startswith("__root__/")
        }

    def restore_layer(self, layer):
        """
        Restore layer for dry-run or in case of failure
        """
        precommit, commit = self.restore[layer["ident"]]
        target = layer["workdir"]
        if commit:
            sp.run(["git", "reset", "--hard", f"{commit}~"], cwd=target, check=True)
        if precommit:
            sp.run(["git", "-reset", f"{precommit}~"], cwd=target, check=True)

    @SubCommand.with_lock
    def run(self):
        "Process given layers"
        self.restore = {}  # Info for restoring for dry-run
        paths = set()
        layers = {layer["ident"]: layer for layer in self.sync.layers if layer["ident"]}
        idents = self.args.ident
        if idents == ["*"]:
            idents = layers.keys()
        for ident in idents:
            assert ident in layers, "Invalid ident"
            paths.update(self.run_layer(layers[ident]))

        if not paths:
            return
        paths = sorted(paths)
        try:
            self._playback_paths(paths)
        except Exception:
            for ident in idents:
                self.restore_layer(layers[ident])
            # if we are not in dryrun we can't be sure we havent already
            # committed some stuff to the data-fs so playback all paths
            if not self.args.dry_run and paths:
                self.sync.playback_paths(
                    paths=paths,
                    recurse=False,
                    override=True,
                    skip_errors=True,
                    dryrun=False,
                )
            raise

        if self.args.dry_run:
            for ident in idents:
                self.restore_layer(layers[ident])
        else:
            self.sync.record(
                paths, recurse=False, skip_errors=True, ignore_removed=True
            )
            for path in paths:
                if self.sync.fs_pathinfo(path)["layeridx"] == 0:
                    self.logger.warning("Conflict with object in custom layer: " + path)
